(* type exp *)
type exp = CoreML.t = Var of Strng.t
                    | Con of Strng.t
                    | App of (exp * exp)
                    | Abs of (Strng.t * exp)
                    | Let of (Strng.t * exp * exp)
                    | Ite of (exp * exp * exp)
(* type typ *)
type typ = TVar of int
         | TFun of (typ * typ)
         | TCon of (Strng.t * typ list)
(* con : string -> exp *)
let con c = Con(Strng.of_string c)
(* tvar : int -> typ *)
let tvar v = TVar v
(* tcon : string -> typ list -> typ *)
let tcon c ts = TCon(Strng.of_string c,ts)
(* tint : typ *)
let tint = tcon "int" []
(* tbool : typ *)
let tbool = tcon "bool" []
(* tlist : typ -> typ *)
let tlist t = tcon "list" [t]
(* tpair : (typ * typ) -> typ *)
let tpair(s,t) = tcon "pair" [s;t]
(* tfun : typ -> typ -> typ *)
let (@->) s t = TFun(s,t)
(* the primitive environment *)
let p = [
 (con "+", tint @-> tint @-> tint);
 (con "-", tint @-> tint @-> tint);
 (con "*", tint @-> tint @-> tint);
 (con "/", tint @-> tint @-> tint);
 (con "<", tvar(-1) @-> tvar(-1)  @-> tbool);
 (con ">", tvar(-2) @-> tvar(-2) @-> tbool);
 (con "=", tvar(-3) @-> tvar(-3) @-> tbool);
 (con "0", tint);
 (con "1", tint);
 (con "2", tint);
 (con "3", tint);
 (con "not", tbool @-> tbool);
 (con "true", tbool);
 (con "false", tbool);
 (con "pair", tpair(tvar(-4),tvar(-5)));
 (con "fst", tpair(tvar(-6),tvar(-7)) @-> tvar(-6));
 (con "snd", tpair(tvar(-8),tvar(-9)) @-> tvar(-9));
 (con "[]", tlist(tvar(-10)));
 (con "nil", tlist(tvar(-10)));
 (con "cons", (tvar(-11)) @-> tlist(tvar(-11)) @-> tlist(tvar(-11)));
 (con "hd", tlist(tvar(-12)) @-> tvar(-12));
 (con "tl", tlist(tvar(-13)) @-> tlist(tvar(-13)));
]
(* the primitive environment extended by `Y' *)
let pmu = [
 (con "Y", (tvar(-14) @-> tvar(-14)) @-> tvar(-14));
] @ p
(* dom : ('a * 'b)list -> 'a list *)
let dom env = List.map fst env
(* lookup : 'a -> 'b -> ('a * 'b)list -> 'b *)
let lookup x d s = match Lst.lookup x s with None   -> d | Some y -> y
(* sub : (int * typ)list -> typ -> typ *)
let rec sub s = function TVar a      -> lookup a (TVar a) s
                       | TFun(t1,t2) -> TFun(sub s t1, sub s t2)
                       | TCon(g,ts)  -> TCon(g,Lst.map (sub s) ts)
(* tvars : typ -> int St.t *)
let rec tvars = function
 | TVar x     -> St.singleton x
 | TFun(s,t)  -> St.union (tvars s) (tvars t)
 | TCon(g,ts) -> Lst.foldr (fun t -> St.union (tvars t)) St.empty ts
(* max_tvar : typ -> int *)
let rec max_tvar = function
 | TVar i      -> i
 | TFun(t1,t2) -> max (max_tvar t1) (max_tvar t2)
 | TCon(_,ts)  -> Lst.foldr (fun t -> max (max_tvar t)) 0 ts
(* to_up_step *)
let to_up_step i (env,e,t) = match Lst.lookup e env with
  | Some t' -> (i,[(t',t)],[])
  | None -> match e with
    | App(e1,e2)    ->
       (i+1,[],[(env,e1,tvar i @-> t);(env,e2,tvar i)])
    | Abs(x,e)      -> 
       (i+2,[(t,tvar i @-> tvar(i+1))],
          [((Var x,tvar i)::env,e,tvar(i+1))])
    | Let(x,e1,e2)  -> 
       (i+1,[],[(env,e1,tvar i);((Var x,tvar i)::env,e2,t)])
    | Ite(e1,e2,e3) -> 
       (i,[],[(env,e1,tbool);(env,e2,t);(env,e3,t)])
    | Var x -> failwith ("unknown `" ^ Strng.to_string x^"'");;
(* warning: the following function is inefficient *)
(* to_up : ((exp * typ)list * exp * typ) -> (typ * typ)list *)
let to_up (env,e,t) =
  let rec to_up (i,eqs) = function
    | [] -> (i,eqs)
    | p::ps ->
       let (i,eqs2,qs) = to_up_step i p in
       let (i,eqs1) = to_up (i,eqs) qs in
       to_up (i,eqs1@eqs2) ps
  in let i = (*largest type variable occurring in t and env*)
    Lst.foldr
      (fun (_, t) -> max (max_tvar t)) (max_tvar t) env in
  snd (to_up (i+1,[]) [(env,e,t)])
(* compose : (int * typ)list -> (int * typ)list -> (int * typ)list *)
let (<*>) sub2 sub1 = (* sub2 after sub1 *)
  let d1 = dom sub1 in
  Lst.map (fun (a,t) -> (a,sub sub2 t)) sub1
    @ Lst.filter (fun (a,_) -> not(Lst.mem a d1)) sub2
(* step : (typ * typ) -> ((typ * typ)list,(int * typ)list) *)
let step = function
  | (s,t) when s = t          -> ([],[])
  | (TVar a,t) | (t,TVar a)   ->
     if St.mem a (tvars t) then failwith "occur check!"
                           else ([],[(a,t)])
  | (TFun(s1,t1),TFun(s2,t2)) -> ([(s1,s2);(t1,t2)],[])
  | (TCon(g,ss),TCon(h,ts))   ->
     if g = h then (Lst.zip ss ts,[])
              else failwith("mismatch: `"^(Strng.to_string g)
                         ^ "' vs. `"^(Strng.to_string h)^"'")
(* unify : (typ * typ)list -> (int * typ)list *)
let unify eqs =
  let rec unify s = function 
    | [] -> s 
    | eq::eqs ->
       let (e,s') = step eq in
       let eqs' =
         Lst.map (fun(l,r) -> (sub s' l,sub s' r)) eqs in
       unify (s' <*> s) (e @ eqs')
  in
  unify [] eqs
(* infer : string -> typ *)
let infer s =
  let e  = CoreML.of_string s in
  let up = to_up(pmu,e,tvar 0) in
  let s  = unify up in
  sub s (tvar 0)
(* toplevel_printer : Format.formatter -> typ -> unit *)
let rec toplevel_printer fmt = function
 | TVar i when i < 0 -> Format.fprintf fmt "b%i" ~-i
 | TVar i -> Format.fprintf fmt "a%i" i
 | TCon(g,[]) -> Format.fprintf fmt "%s" (Strng.to_string g)
 | TCon(g,ts) -> Format.fprintf fmt "%s(%a)" (Strng.to_string g)
  toplevel_printer_list ts
 | TFun(TFun _ as t1,t2) -> Format.fprintf fmt "(%a) -> %a"
  toplevel_printer t1
  toplevel_printer t2
 | TFun(t1,t2) -> Format.fprintf fmt "%a -> %a"
  toplevel_printer t1
  toplevel_printer t2
and toplevel_printer_list fmt = function
 | [] -> ()
 | [t] -> toplevel_printer fmt t
 | t::ts -> Format.fprintf fmt "%a,%a" toplevel_printer t toplevel_printer_list ts
(* exemplary unification problems *)
let up1 = [
 (tlist(tbool),tlist(tvar 0))
]

let up2 = [
 (tvar 1,tvar 6 @-> tvar 5);
 (tvar 4,tvar 7 @-> tvar 6);
 (tvar 4,tvar 7);
 (tvar 3 @-> tvar 2,tvar 4 @-> tvar 5);
 (tvar 1,tvar 10 @-> tvar 9);
 (tvar 8,tvar 11 @-> tvar 10);
 (tvar 8,tvar 11);
 (tvar 3,tvar 8 @-> tvar 9);
 (tvar 0,tvar 1 @-> tvar 2);
]

let up3 = [
 (tvar 2,tvar 3);
 (tvar 1,tvar 2 @-> tvar 3);
 (tvar 1,tvar 4 @-> tvar 0);
 (tint,tvar 4);
]

let up4 = [
 (tvar 1,tvar 3 @-> tvar 2);
 (tvar 1,tvar 3);
 (tvar 0,tvar 1 @-> tvar 2);
]
